@@ -19,7 +19,7 @@
    You should have received a copy of the GNU Library General Public
    License along with the Gnome Library; see the file COPYING.LIB.  If not,
    write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-   Boston, MA 02111-1307, USA.  
+   Boston, MA 02111-1307, USA.
 
    Authors:
    Miguel De Icaza <miguel@helixcode.com>
@@ -46,33 +46,60 @@
 #include "gnomesupport.h"
 #endif
 
+#define FAST_FILE_EOF -1
+#define FAST_FILE_BUFFER_SIZE (1024 * 16)
 
-#if !defined getc_unlocked && !defined HAVE_GETC_UNLOCKED
-# define getc_unlocked(fp) getc (fp)
-#endif
+typedef struct {
+	guchar *ptr;
+	guchar *buffer;
+	int     length;
+	FILE   *fh;
+} FastFile;
 
-
-/* The only goal of this function is to make sure no comment line is ever returned
-   to the 2 parsers below.
-   It is evil because I could not figure out what was wrong with those parsers.
-   They should ignore all comments but they do not ignore them. They insert them 
-   in the hash table. This function makes sure this never ever happens.
-   -- Mathieu - who takes all responsbility for this complete evilness.
-*/
-static int 
-hack_getc (FILE *stream) 
+static void
+fast_file_close (FastFile *ffh)
 {
-	static int previous_char = '\n';
-	int current_char;
+	fclose (ffh->fh);
+	g_free (ffh->buffer);
+}
 
-	current_char = getc_unlocked (stream);
+static gboolean
+fast_file_open (FastFile *ffh, const char *filename)
+{
+	memset (ffh, 0, sizeof (FastFile));
 
-	if (current_char == '#' && 
-	    previous_char == '\n') {
-		while (getc_unlocked (stream) != '\n') {}
-		return hack_getc (stream);
-	} else {
-		return current_char;
+	ffh->fh = fopen (filename, "r");
+	if (ffh->fh == NULL) {
+		return FALSE;
+	}
+
+	ffh->buffer = g_malloc (FAST_FILE_BUFFER_SIZE);
+	ffh->ptr = ffh->buffer;
+
+	ffh->length = fread (ffh->buffer, 1, FAST_FILE_BUFFER_SIZE, ffh->fh);
+
+	if (ffh->length < 0) {
+		fast_file_close (ffh);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static inline int
+fast_file_getc (FastFile *ffh)
+{
+	if (ffh->ptr < ffh->buffer + ffh->length)
+		return *(ffh->ptr++);
+	else {
+		ffh->length = fread (ffh->buffer, 1, FAST_FILE_BUFFER_SIZE, ffh->fh);
+
+		if (!ffh->length)
+			return FAST_FILE_EOF;
+		else {
+			ffh->ptr = ffh->buffer;
+			return *(ffh->ptr++);
+		}
 	}
 }
 
@@ -106,12 +133,8 @@ static int should_write_file_back = 0;
 static GList *current_lang = NULL;
 /* we want to replace the previous key if the current key has a higher
    language level */
-static char *previous_key = NULL;
 static int previous_key_lang_level = -1;
 
-
-
-
 /*
  * A hash table containing all of the Mime records for specific
  * mime types (full description, like image/png)
@@ -138,11 +161,10 @@ static GHashTable *registered_types_user;
 static void	      reload_if_needed (void);
 static GnomeVFSResult write_back_mime_user_file (void);
 static GnomeVFSResult write_back_keys_user_file (void);
-static const char *   gnome_vfs_mime_get_registered_mime_type_key (const char *mime_type, 
+static const char *   gnome_vfs_mime_get_registered_mime_type_key (const char *mime_type,
 								   const char *key);
 
-
-static gboolean 
+static gboolean
 does_string_contain_caps (const char *string)
 {
 	const char *temp_c;
@@ -158,8 +180,6 @@ does_string_contain_caps (const char *string)
 	return FALSE;
 }
 
-
-
 static GnomeMimeContext *
 context_new (GHashTable *hash_table, GString *str)
 {
@@ -180,31 +200,24 @@ context_new (GHashTable *hash_table, GString *str)
 		g_free (mime_type);
 		return context;
 	}
-	
+
+/*	fprintf (stderr, "New context: '%s'\n", mime_type); */
+
 	context = g_new (GnomeMimeContext, 1);
 	context->mime_type = mime_type;
-	context->keys = g_hash_table_new (g_str_hash, g_str_equal);
+	context->keys = g_hash_table_new_full (
+		g_str_hash, g_str_equal,
+		(GDestroyNotify) g_free,
+		(GDestroyNotify) g_free);
 
 	g_hash_table_insert (hash_table, context->mime_type, context);
+
 	return context;
 }
 
-static gboolean
-release_key_and_value (gpointer key, gpointer value, gpointer user_data)
-{
-	g_free (key);
-	g_free (value);
-
-	return TRUE;
-}
-
 static void
 context_destroy (GnomeMimeContext *context)
 {
-	/*
-	 * Destroy it
-	 */
-	g_hash_table_foreach_remove (context->keys, release_key_and_value, NULL);
 	g_hash_table_destroy (context->keys);
 	g_free (context->mime_type);
 	g_free (context);
@@ -249,51 +262,23 @@ language_level (const char *langage)
 static void
 context_add_key (GnomeMimeContext *context, char *key, char *lang, char *value)
 {
-	char *v;
-	char *orig_key;
 	int lang_level;
 
-	lang_level = language_level(lang);
+	lang_level = language_level (lang);
 	/* wrong language completely */
 	if (lang_level < 0)
 		return;
 
-	/* if we have some language defined and
-	   if there was a previous_key */
-	if (lang_level > 0 && previous_key) {
-		/* if our new key has a better lang_level then remove the
-		   previous key */
-		if (previous_key_lang_level <= lang_level) {
-			if (g_hash_table_lookup_extended (context->keys,
-							  previous_key,
-							  (gpointer *)&orig_key,
-							  (gpointer *)&v)) {
-				g_hash_table_remove (context->keys, orig_key);
-				g_free(orig_key);
-				g_free(v);
-			}
-		/* else, our language level really sucks and the previous
-		   translation was of better language quality so just
-		   ignore us */
-		} else {
-			return;
-		}
+	/* if a previous key in the hash had a better lang_level don't do anything */
+	if (lang_level > 0 &&
+	    previous_key_lang_level > lang_level) {
+		return;
 	}
 
-	if (g_hash_table_lookup_extended (context->keys, key,
-					  (gpointer *)&orig_key,
-					  (gpointer *)&v)) {
-		/* if we found it in the database already, just replace it here */
-		g_free (v);
-		g_hash_table_insert (context->keys, orig_key,
-				     g_strdup (value));
-	} else {
-		g_hash_table_insert (context->keys, g_strdup(key),
-				     g_strdup (value));
-	}
-	/* set this as the previous key */
-	g_free(previous_key);
-	previous_key = g_strdup(key);
+/*	fprintf (stderr, "Add key: '%s' '%s' '%s' %d\n", key, lang, value, lang_level); */
+
+	g_hash_table_replace (context->keys, g_strdup (key), g_strdup (value));
+
 	previous_key_lang_level = lang_level;
 }
 
@@ -306,354 +291,210 @@ typedef enum {
 	STATE_ON_VALUE
 } ParserState;
 
+#define APPEND_CHAR(gstr,c) g_string_insert_c ((gstr), -1, (c))
+
+typedef enum {
+	FORMAT_MIME,
+	FORMAT_KEYS
+} Format;
+
+#define load_mime_type_info_from(a,b) load_type_info_from ((a), (b), FORMAT_MIME)
+#define load_mime_list_info_from(a,b) load_type_info_from ((a), (b), FORMAT_KEYS)
+
 static void
-load_mime_type_info_from (const char *filename, GHashTable *hash_table)
+load_type_info_from (const char *filename,
+		     GHashTable *hash_table,
+		     Format      format)
 {
-	FILE *mime_file;
-	gboolean in_comment, context_used;
 	GString *line;
 	int column, c;
 	ParserState state;
+	FastFile mime_file;
+	gboolean skip_line;
 	GnomeMimeContext *context;
-	char *key;
-	char *lang;
-	
-	mime_file = fopen (filename, "r");
-	if (mime_file == NULL)
-		return;
+	int key, lang, last_str_end; /* offsets */
 
-	in_comment = FALSE;
-	context_used = FALSE;
+	if (!fast_file_open (&mime_file, filename)) {
+		return;
+	}
+
+	skip_line = FALSE;
 	column = -1;
 	context = NULL;
-	key = NULL;
-	lang = NULL;
 	line = g_string_sized_new (120);
+	key = lang = last_str_end = 0;
 	state = STATE_NONE;
-	
-	while ((c = hack_getc (mime_file)) != EOF){
+
+	while ((c = fast_file_getc (&mime_file)) != EOF) {
+	handle_char:
 		column++;
 		if (c == '\r')
 			continue;
 
-		if (c == '#' && column == 0){		
-			in_comment = TRUE;
+		if (c == '#' && column == 0) {
+			skip_line = TRUE;
 			continue;
 		}
-		
-		if (c == '\n'){
-			in_comment = FALSE;
+
+		if (skip_line) {
+			if (c == '\n') {
+				skip_line = FALSE;
+				column = -1;
+				g_string_assign (line, "");
+				key = lang = last_str_end = 0;
+			}
+			continue;
+		}
+
+		if (c == '\n') {
+			skip_line = FALSE;
 			column = -1;
-			if (state == STATE_ON_MIME_TYPE){
-
-				/* set previous key to nothing
-				   for this mime type */
-				g_free(previous_key);
-				previous_key = NULL;
+			if (state == STATE_ON_MIME_TYPE) {
+				/* setup for a new key */
 				previous_key_lang_level = -1;
-
 				context = context_new (hash_table, line);
-				context_used = FALSE;
-				g_string_assign (line, "");
-				state = STATE_LOOKING_FOR_KEY;
-				continue;
-			}
-			if (state == STATE_ON_VALUE){
-				context_used = TRUE;
-				context_add_key (context, key, lang, line->str);
-				g_string_assign (line, "");
-				g_free (key);
-				key = NULL;
-				g_free (lang);
-				lang = NULL;
-				state = STATE_LOOKING_FOR_KEY;
-				continue;
+
+			} else if (state == STATE_ON_VALUE) {
+				APPEND_CHAR (line, '\0');
+				context_add_key (context,
+						 line->str + key,
+						 lang ? line->str + lang : NULL,
+						 line->str + last_str_end);
+				key = lang = 0;
 			}
+			g_string_assign (line, "");
+			last_str_end = 0;
+			state = STATE_LOOKING_FOR_KEY;
 			continue;
 		}
 
-		if (in_comment)
-			continue;
-
-		switch (state){
+		switch (state) {
 		case STATE_NONE:
-			if (c != ' ' && c != '\t')
-				state = STATE_ON_MIME_TYPE;
-			else
+			if (c == ' ' || c == '\t') {
 				break;
+			} else if (c == ':') {
+				skip_line = TRUE;
+				break;
+			} else {
+				state = STATE_ON_MIME_TYPE;
+			}
 			/* fall down */
-			
+
 		case STATE_ON_MIME_TYPE:
-			if (c == ':'){
-				in_comment = TRUE;
-				break;
-			}
-			g_string_append_c (line, c);
-			break;
-
-		case STATE_LOOKING_FOR_KEY:
-			if (c == '\t' || c == ' ')
-				break;
-
-			if (c == '['){
-				state = STATE_LANG;
-				break;
-			}
-
-			if (column == 0){
-				state = STATE_ON_MIME_TYPE;
-				g_string_append_c (line, c);
-				break;
-			}
-			state = STATE_ON_KEY;
-			/* falldown */
-
-		case STATE_ON_KEY:
-			if (c == '\\'){
-				c = hack_getc (mime_file);
-				if (c == EOF)
-					break;
-			}
-			if (c == '='){
-				key = g_strdup (line->str);
-				g_string_assign (line, "");
-				state = STATE_ON_VALUE;
-				break;
-			}
-			g_string_append_c (line, c);
-			break;
-
-		case STATE_ON_VALUE:
-			g_string_append_c (line, c);
-			break;
-			
-		case STATE_LANG:
-			if (c == ']'){
-				state = STATE_ON_KEY;      
-				if (line->str [0]){
-					g_free(lang);
-					lang = g_strdup(line->str);
-				} else {
-					in_comment = TRUE;
-					state = STATE_LOOKING_FOR_KEY;
-				}
-				g_string_assign (line, "");
-				break;
-			}
-			g_string_append_c (line, c);
-			break;
-		}
-	}
-
-	if (context != NULL) {
-		if (key && line->str [0])
-			context_add_key (context, key, lang, line->str);
-		else
-			if (!context_used)
-				context_destroy_and_unlink (context);
-	}
-
-	g_string_free (line, TRUE);
-	g_free (key);
-	g_free (lang);
-
-	/* free the previous_key stuff */
-	g_free(previous_key);
-	previous_key = NULL;
-	previous_key_lang_level = -1;
-
-	fclose (mime_file);
-}
-
-/*
- *  load_mime_list_info_from
- *
- *  Why this special function when a similar one is already in the code?
- *  Because we need to handle the case where ':' is used to delimit
- *  the start of a key in a .mime file instead of '=' as is used in
- *  the .key file.  Why is this done?  Why are there two mime database 
- *  files with differing standards?  We may never know.  
- *  Until we have a better solution, this will suffice.	
- *  
- *  Both ':' and '=' are used to delimit the start of a key.
- */
- 
-static void
-load_mime_list_info_from (const char *filename, GHashTable *hash_table)
-{
-	FILE *mime_file;
-	gboolean in_comment, context_used;
-	GString *line;
-	int column, c;
-	ParserState state;
-	GnomeMimeContext *context;
-	char *key;
-	char *lang;
-	
-	mime_file = fopen (filename, "r");
-	if (mime_file == NULL)
-		return;
-
-	in_comment = FALSE;
-	context_used = FALSE;
-	column = -1;
-	context = NULL;
-	key = NULL;
-	lang = NULL;
-	line = g_string_sized_new (120);
-	state = STATE_NONE;
-	
-	while ((c = hack_getc (mime_file)) != EOF){
-		column++;
-		if (c == '\r')
-			continue;
-
-		if (c == '#' && column == 0){
-			in_comment = TRUE;
-			continue;
-		}
-		
-		if (c == '\n'){
-			in_comment = FALSE;
-			column = 0;
-			if (state == STATE_ON_MIME_TYPE){
-				/* set previous key to nothing
-				   for this mime type */
-				g_free(previous_key);
-				previous_key = NULL;
-				previous_key_lang_level = -1;
-
-				context = context_new (hash_table, line);
-				context_used = FALSE;
-				g_string_assign (line, "");
-				state = STATE_LOOKING_FOR_KEY;
-				continue;
-			}
-			if (state == STATE_ON_VALUE){
-				context_used = TRUE;
-				context_add_key (context, key, lang, line->str);
-				g_string_assign (line, "");
-				g_free (key);
-				key = NULL;
-				g_free (lang);
-				lang = NULL;
-				state = STATE_LOOKING_FOR_KEY;
-				continue;
-			}
-			continue;
-		}
-
-		if (in_comment) {
-			continue;
-		}
-
-		switch (state){
-		case STATE_NONE:
-			if (c != ' ' && c != '\t')
-				state = STATE_ON_MIME_TYPE;
-			else
-				break;
-			/* fall down */
-			
-		case STATE_ON_MIME_TYPE:
-			if (c == ':'){
-				in_comment = TRUE;
-				break;
-			}
-			g_string_append_c (line, c);
-			break;
-
-		case STATE_LOOKING_FOR_KEY:
-			if (c == '\t' || c == ' ')
-				break;
-
-			if (c == '['){
-				state = STATE_LANG;
-				break;
-			}
-
-			if (column == 1){
-				state = STATE_ON_MIME_TYPE;
-				g_string_append_c (line, c);
-				break;
-			}
-			state = STATE_ON_KEY;
-			/* falldown */
-
-		case STATE_ON_KEY:
-			if (c == '\\'){
-				c = hack_getc (mime_file);
-				if (c == EOF)
-					break;
-			}			
-			if (c == '=') {
-				key = g_strdup (line->str);
-				g_string_assign (line, "");
-				state = STATE_ON_VALUE;
-				break;
-			}
-
 			if (c == ':') {
-				key = g_strdup (line->str);				
-				g_string_assign (line, "");
+				skip_line = TRUE;
+				/* setup for a new key */
+				previous_key_lang_level = -1;
+				context = context_new (hash_table, line);
+				state = STATE_LOOKING_FOR_KEY;
+				break;
+			}
+			APPEND_CHAR (line, c);
+			break;
+
+		case STATE_LOOKING_FOR_KEY:
+			if (c == '\t' || c == ' ') {
+				break;
+			}
+
+			if (c == '[') {
+				state = STATE_LANG;
+				break;
+			}
+
+			if (column == 0) {
+				state = STATE_ON_MIME_TYPE;
+				APPEND_CHAR (line, c);
+				break;
+			}
+			state = STATE_ON_KEY;
+			/* falldown */
+
+		case STATE_ON_KEY:
+			if (c == '\\') {
+				c = fast_file_getc (&mime_file);
+				if (c == EOF) {
+					break;
+				}
+			}
+			if (c == '=') {
+				key = last_str_end;
+				APPEND_CHAR (line, '\0');
+				last_str_end = line->len;
+				state = STATE_ON_VALUE;
+				break;
+			}
+
+			if (format == FORMAT_KEYS && c == ':') {
+				key = last_str_end;
+				APPEND_CHAR (line, '\0');
+				last_str_end = line->len;
 
 				/* Skip space after colon.  There should be one
 				 * there.  That is how the file is defined. */
-				c = hack_getc (mime_file);
+				c = fast_file_getc (&mime_file);
 				if (c != ' ') {
-					/* Revert seek */
-					ungetc (c, mime_file);
+					if (c == EOF) {
+						break;
+					} else {
+						goto handle_char;
+					}
 				} else {
 					column++;
 				}
-				
+
 				state = STATE_ON_VALUE;
 				break;
 			}
 
-			g_string_append_c (line, c);
+			APPEND_CHAR (line, c);
 			break;
 
 		case STATE_ON_VALUE:
-			g_string_append_c (line, c);
+			APPEND_CHAR (line, c);
 			break;
-			
+
 		case STATE_LANG:
 			if (c == ']') {
-				state = STATE_ON_KEY;      
-				if (line->str [0]){
-					g_free(lang);
-					lang = g_strdup(line->str);
-				} else {
-					in_comment = TRUE;
+				state = STATE_ON_KEY;
+
+				lang = last_str_end;
+				APPEND_CHAR (line, '\0');
+				last_str_end = line->len;
+
+				if (!line->str [0] ||
+				    language_level (line->str + lang) < 0) {
+					skip_line = TRUE;
+					key = lang = last_str_end = 0;
+					g_string_assign (line, "");
 					state = STATE_LOOKING_FOR_KEY;
 				}
-				g_string_assign (line, "");
-				break;
+			} else {
+				APPEND_CHAR (line, c);
 			}
-			g_string_append_c (line, c);
 			break;
 		}
 	}
 
 	if (context != NULL) {
-		if (key && line->str [0])
-			context_add_key (context, key, lang, line->str);
-		else
-			if (!context_used)
+		if (key && line->str [0]) {
+			APPEND_CHAR (line, '\0');
+			context_add_key (context,
+					 line->str + key,
+					 lang ? line->str + lang : NULL,
+					 line->str + last_str_end);
+		} else {
+			if (g_hash_table_size (context->keys) < 1) {
 				context_destroy_and_unlink (context);
+			}
+		}
 	}
 
 	g_string_free (line, TRUE);
-	g_free (key);
-	g_free (lang);
 
-	/* free the previous_key stuff */
-	g_free(previous_key);
-	previous_key = NULL;
 	previous_key_lang_level = -1;
 
-	fclose (mime_file);
+	fast_file_close (&mime_file);
 }
 
 static void
@@ -663,25 +504,25 @@ mime_info_load (mime_dir_source_t *source)
 	struct dirent *dent;
 	const int extlen = sizeof (".keys") - 1;
 	char *filename;
-	
+
 	if (stat (source->dirname, &source->s) != -1)
 		source->valid = TRUE;
 	else
 		source->valid = FALSE;
-	
+
 	dir = opendir (source->dirname);
-	if (!dir){
+	if (!dir) {
 		source->valid = FALSE;
 		return;
 	}
-	if (source->system_dir){
+	if (source->system_dir) {
 		filename = g_strconcat (source->dirname, "/gnome-vfs.keys", NULL);
 		load_mime_type_info_from (filename, specific_types);
 		g_free (filename);
 	}
 
-	while ((dent = readdir (dir)) != NULL){
-		
+	while ((dent = readdir (dir)) != NULL) {
+
 		int len = strlen (dent->d_name);
 
 		if (len <= extlen)
@@ -697,7 +538,7 @@ mime_info_load (mime_dir_source_t *source)
 			 */
 			continue;
 		}
-			
+
 		if (!source->system_dir && !strcmp (dent->d_name, "user.keys"))
 			continue;
 
@@ -720,25 +561,25 @@ mime_list_load (mime_dir_source_t *source)
 	struct dirent *dent;
 	const int extlen = sizeof (".mime") - 1;
 	char *filename;
-	
+
 	if (stat (source->dirname, &source->s) != -1)
 		source->valid = TRUE;
 	else
 		source->valid = FALSE;
-	
+
 	dir = opendir (source->dirname);
-	if (!dir){
+	if (!dir) {
 		source->valid = FALSE;
 		return;
 	}
-	if (source->system_dir){
+	if (source->system_dir) {
 		filename = g_strconcat (source->dirname, "/gnome-vfs.mime", NULL);
 		load_mime_list_info_from (filename, registered_types);
 		g_free (filename);
 	}
 
-	while ((dent = readdir (dir)) != NULL){
-		
+	while ((dent = readdir (dir)) != NULL) {
+
 		int len = strlen (dent->d_name);
 
 		if (len <= extlen)
@@ -754,7 +595,7 @@ mime_list_load (mime_dir_source_t *source)
 			 */
 			continue;
 		}
-			
+
 		if (!source->system_dir && !strcmp (dent->d_name, "user.mime"))
 			continue;
 
@@ -790,17 +631,17 @@ gnome_vfs_mime_init (void)
 
 	specific_types_user = g_hash_table_new (g_str_hash, g_str_equal);
 	registered_types_user  = g_hash_table_new (g_str_hash, g_str_equal);
-	
+
 	current_lang = gnome_vfs_i18n_get_language_list ("LC_MESSAGES");
 
 	/*
 	 * Setup the descriptors for the information loading
 	 */
-
 	gnome_mime_dir.dirname = g_strdup (DATADIR "/mime-info");
 	gnome_mime_dir.system_dir = TRUE;
-	
-	user_mime_dir.dirname  = g_strconcat (g_get_home_dir(), "/.gnome/mime-info", NULL);
+
+	user_mime_dir.dirname  = g_strconcat
+		(g_get_home_dir (), "/.gnome/mime-info", NULL);
 	user_mime_dir.system_dir = FALSE;
 
 	/*
@@ -818,7 +659,7 @@ remove_keys (gpointer key, gpointer value, gpointer user_data)
 	GnomeMimeContext *context = value;
 
 	context_destroy (context);
-	
+
 	return TRUE;
 }
 
@@ -828,7 +669,7 @@ reload_if_needed (void)
 	time_t now = time (NULL);
 	gboolean need_reload = FALSE;
 	struct stat s;
-	
+
 	if (last_checked + 5 >= now)
 		return;
 
@@ -841,7 +682,7 @@ reload_if_needed (void)
 			need_reload = TRUE;
 
 	last_checked = now;
-	
+
 	if (need_reload) {
 	        gnome_vfs_mime_info_reload ();
 	}
@@ -896,21 +737,11 @@ gnome_vfs_mime_info_reload (void)
 
 	/* 1. Clean */
 	gnome_vfs_mime_info_clear ();
-	
+
 	/* 2. Reload */
 	load_mime_type_info ();
 
 	/* 3. Tell anyone who cares */
-	/* FIXME bugzilla.eazel.com 5459:
-	 * This is called only when some client asks for data, so changes made
-	 * to the MIME data via (e.g.) the File Types and Programs capplet
-	 * won't be reflected in clients (e.g. Nautilus) until the next time
-	 * some client asks for MIME data. One way to fix this is to implement
-	 * the gconf solution mentioned in bug 5460. Another possibility is to
-	 * use file-node monitoring to notice when the contents of the data
-	 * directory have changed, but file-node monitoring is only a pipe
-	 * dream at the moment.
-	 */
 	gnome_vfs_mime_monitor_emit_data_changed (gnome_vfs_mime_monitor_get ());
 }
 
@@ -918,7 +749,7 @@ gnome_vfs_mime_info_reload (void)
 /**
  * gnome_vfs_mime_freeze
  *
- * Freezes the mime data so that you can do multiple 
+ * Freezes the mime data so that you can do multiple
  * updates to the dat in one batch without needing
  * to back the files to disk or readind them
  */
@@ -933,7 +764,7 @@ gnome_vfs_mime_freeze (void)
 /**
  * gnome_vfs_mime_thaw
  *
- * UnFreezes the mime data so that you can do multiple 
+ * UnFreezes the mime data so that you can do multiple
  * updates to the dat in one batch without needing
  * to back the files to disk or readind them
  */
@@ -944,40 +775,32 @@ gnome_vfs_mime_thaw (void)
 
 	if (should_write_file_back == 0) {
 		write_back_mime_user_file ();
-		write_back_keys_user_file ();		
+		write_back_keys_user_file ();
 	}
 }
 
 
 static GnomeVFSResult
-set_value_real (const char *mime_type, const char *key, const char *value, 
+set_value_real (const char *mime_type, const char *key, const char *value,
 		GHashTable *user_hash_table)
 {
 	GnomeMimeContext *context;
 
-	if (mime_type == NULL 
+	if (mime_type == NULL
 	    || key == NULL
 	    || value == NULL) {
 		return gnome_vfs_result_from_errno ();
 	}
 
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      gnome_vfs_result_from_errno ());
-	
+
 	if (!gnome_vfs_mime_inited) {
 		gnome_vfs_mime_init ();
 	}
 
 	context = g_hash_table_lookup (user_hash_table, mime_type);
-	if (context != NULL) {
-		gpointer orig_key, orig_value;
-		if (g_hash_table_lookup_extended (context->keys, key, &orig_key, &orig_value)) {
-			g_hash_table_insert (context->keys, orig_key, g_strdup (value));
-			g_free (orig_value);
-		} else {
-			g_hash_table_insert (context->keys, g_strdup (key), g_strdup (value));
-		}
-	} else {
+	if (context == NULL) {
 		GString *string;
 
 		string = g_string_new (mime_type);
@@ -986,8 +809,9 @@ set_value_real (const char *mime_type, const char *key, const char *value,
 		context = context_new (user_hash_table, string);
 		/* add the info to the mime type context */
 		g_hash_table_insert (context->keys, g_strdup (key), g_strdup (value));
-	}
-	
+	} else
+		g_hash_table_replace (context->keys, g_strdup (key), g_strdup (value));
+
 	return GNOME_VFS_OK;
 }
 
@@ -997,10 +821,10 @@ set_value_real (const char *mime_type, const char *key, const char *value,
  * @key: a key to store the value in.
  * @value: the value to store in the key.
  *
- * This function is going to set the value 
+ * This function is going to set the value
  * associated to the key and it will save it
  * to the user' file if necessary.
- * You should not free the key/values passed to 
+ * You should not free the key/values passed to
  * this function. They are used internally.
  *
  * Returns: GNOME_VFS_OK if the operation succeeded, otherwise an error code
@@ -1010,9 +834,9 @@ GnomeVFSResult
 gnome_vfs_mime_set_value (const char *mime_type, const char *key, const char *value)
 {
 	GnomeVFSResult retval;
-	
+
 	retval = set_value_real (mime_type, key, value, specific_types_user);
-	
+
 	if (should_write_file_back == 0) {
 		return write_back_keys_user_file ();
 	}
@@ -1045,14 +869,14 @@ get_value_from_hash_table (GHashTable *hash_table, const char *mime_type, const
 }
 
 static const char *
-get_value_real (const char *mime_type, 
-		const char *key, 
-		GHashTable *user_hash_table, 
+get_value_real (const char *mime_type,
+		const char *key,
+		GHashTable *user_hash_table,
 		GHashTable *system_hash_table)
 {
 	const char *value;
 	char *generic_type, *p;
-	
+
 	g_return_val_if_fail (key != NULL, NULL);
 	g_assert (user_hash_table != NULL);
 	g_assert (system_hash_table != NULL);
@@ -1060,8 +884,8 @@ get_value_real (const char *mime_type,
 	if (mime_type == NULL) {
 		return NULL;
 	}
-	
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      NULL);
 
 	reload_if_needed ();
@@ -1074,23 +898,23 @@ get_value_real (const char *mime_type,
 	if (value != NULL) {
 		return value;
 	}
-	
+
 	value = get_value_from_hash_table (system_hash_table, mime_type, key);
 	if (value != NULL) {
 		return value;
 	}
-	
+
 	generic_type = g_strdup (mime_type);
 	p = strchr (generic_type, '/');
 	if (p != NULL)
 		*(p+1) = '\0';
-	
+
 	value = get_value_from_hash_table (user_hash_table, generic_type, key);
 	if (value != NULL) {
 		g_free (generic_type);
 		return value;
 	}
-	
+
 	value = get_value_from_hash_table (system_hash_table, generic_type, key);
 	g_free (generic_type);
 	if (value != NULL) {
@@ -1106,7 +930,7 @@ get_value_real (const char *mime_type,
  * @mime_type: a mime type.
  * @key: A key to lookup for the given mime-type
  *
- * This function retrieves the value associated with @key in 
+ * This function retrieves the value associated with @key in
  * the given GnomeMimeContext.  The string is private, you
  * should not free the result.
  *
@@ -1131,12 +955,12 @@ gnome_vfs_mime_get_value (const char *mime_type, const char *key)
  */
 gboolean
 gnome_vfs_mime_type_is_known (const char *mime_type)
-{	
+{
 	if (mime_type == NULL) {
 		return FALSE;
 	}
 
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      FALSE);
 
 	if (!gnome_vfs_mime_inited)
@@ -1147,15 +971,15 @@ gnome_vfs_mime_type_is_known (const char *mime_type)
 	if (g_hash_table_lookup (specific_types, mime_type)) {
 		return TRUE;
 	}
-	
+
 	if (g_hash_table_lookup (specific_types_user, mime_type)) {
 		return TRUE;
 	}
-	
+
 	if (g_hash_table_lookup (registered_types, mime_type)) {
 		return TRUE;
 	}
-	
+
 	if (g_hash_table_lookup (registered_types_user, mime_type)) {
 		return TRUE;
 	}
@@ -1174,8 +998,8 @@ void
 gnome_vfs_mime_keys_list_free (GList *mime_type_list)
 {
 	/* we do not need to free the data in the list since
-	   the data was stolen from the internal hash table 
-	   This function is there so that people do not need 
+	   the data was stolen from the internal hash table
+	   This function is there so that people do not need
 	   to know this particuliar implementation detail.
 	*/
 
@@ -1195,7 +1019,7 @@ assemble_list (gpointer key, gpointer value, gpointer user_data)
  * @mime_type: the mime type to lookup.
  *
  * Returns a GList that contains private strings with all of the keys
- * associated with the @mime_type.  
+ * associated with the @mime_type.
  */
 GList *
 gnome_vfs_mime_get_key_list (const char *mime_type)
@@ -1203,29 +1027,29 @@ gnome_vfs_mime_get_key_list (const char *mime_type)
 	char *p, *generic_type;
 	GnomeMimeContext *context;
 	GList *list = NULL, *l;
-	
+
 	if (mime_type == NULL) {
 		return NULL;
 	}
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      NULL);
 
 	if (!gnome_vfs_mime_inited)
 		gnome_vfs_mime_init ();
 
 	reload_if_needed ();
-	
+
 	generic_type = g_strdup (mime_type);
 	p = strchr (generic_type, '/');
 	if (p != NULL)
 		*(p+1) = 0;
-	
+
 	context = g_hash_table_lookup (specific_types_user, generic_type);
 	if (context != NULL) {
 		g_hash_table_foreach (
 			context->keys, assemble_list, &list);
 	}
-	
+
 	context = g_hash_table_lookup (specific_types, generic_type);
 	if (context != NULL) {
 		g_hash_table_foreach (
@@ -1233,12 +1057,12 @@ gnome_vfs_mime_get_key_list (const char *mime_type)
 	}
 
 	g_free (generic_type);
-	for (l = list; l;){
-		if (l->next){
+	for (l = list; l;) {
+		if (l->next) {
 			void *this = l->data;
 			GList *m;
 
-			for (m = l->next; m; m = m->next){
+			for (m = l->next; m; m = m->next) {
 				if (strcmp ((char*) this, (char*) m->data) != 0)
 					continue;
 				list = g_list_remove (list, m->data);
@@ -1264,7 +1088,7 @@ str_cmp_callback  (gconstpointer a,
 /**
  * gnome_vfs_mime_set_extensions_list:
  * @mime_type: the mime type.
- * @extensions_list: a whitespace-separated list of the 
+ * @extensions_list: a whitespace-separated list of the
  *                   extensions to set for this mime type.
  *
  * Sets the extensions for a given mime type. Overrides
@@ -1295,11 +1119,11 @@ gnome_vfs_mime_get_extensions_list (const char *mime_type)
 	gchar **elements;
 	int index;
 	GnomeMimeContext *context;
-	
+
 	if (mime_type == NULL) {
 		return NULL;
 	}
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      NULL);
 
 
@@ -1334,18 +1158,18 @@ gnome_vfs_mime_get_extensions_list (const char *mime_type)
 	list = NULL;
 	if (extensions != NULL) {
 		/* Parse the extensions and add to list */
-		elements = g_strsplit (extensions, " ", 0);	
+		elements = g_strsplit (extensions, " ", 0);
 		if (elements != NULL) {
 			index = 0;
-			
+
 			while (elements[index] != NULL) {
 				if (strcmp (elements[index], "") != 0) {
 					list = g_list_append (list, g_strdup (elements[index]));
 				}
 				index++;
-			}			
+			}
 			g_strfreev (elements);
-		}		
+		}
 	}
 
 	g_free (extensions);
@@ -1365,11 +1189,11 @@ gnome_vfs_mime_get_extensions_string (const char *mime_type)
 {
 	GList *extensions_list, *temp_list;
 	char *extensions;
-	
+
 	if (mime_type == NULL) {
 		return NULL;
 	}
-	g_return_val_if_fail (!does_string_contain_caps (mime_type), 
+	g_return_val_if_fail (!does_string_contain_caps (mime_type),
 			      NULL);
 
 
@@ -1393,7 +1217,7 @@ gnome_vfs_mime_get_extensions_string (const char *mime_type)
 	extensions[strlen (extensions) - 1] = '\0';
 	return extensions;
 }
-	
+
 /**
  * gnome_vfs_mime_get_extensions_pretty_string:
  * @mime_type: the mime type
@@ -1405,7 +1229,7 @@ gnome_vfs_mime_get_extensions_pretty_string (const char *mime_type)
 {
 	GList *extensions, *element;
 	char *ext_str, *tmp_str;
-	
+
 	if (mime_type == NULL) {
 		return NULL;
 	}
@@ -1418,7 +1242,7 @@ gnome_vfs_mime_get_extensions_pretty_string (const char *mime_type)
 	}
 
 	reload_if_needed ();
-	
+
 	extensions = gnome_vfs_mime_get_extensions_list (mime_type);
 	if (extensions == NULL) {
 		return NULL;
@@ -1427,7 +1251,7 @@ gnome_vfs_mime_get_extensions_pretty_string (const char *mime_type)
 	for (element = extensions; element != NULL; element = element->next) {
 		if (ext_str != NULL) {
 			tmp_str = ext_str;
-			
+
 			if (element->next == NULL) {
 				ext_str = g_strconcat (tmp_str, ".", (char *)element->data, NULL);
 			} else {
@@ -1478,20 +1302,20 @@ mime_list_sort (gconstpointer a, gconstpointer b)
  *
  *  Hash table function that adds the name of the mime type
  *  to the supplied GList.
- *  
+ *
  */
-static void 
+static void
 get_key_name (gpointer key, gpointer value, gpointer user_data)
 {
 	GnomeMimeContext *context;
 	char *name;
 	GList **list = user_data;
 	GList *duplicate;
-	
+
 	if (value == NULL || key == NULL) {
 		return;
 	}
-	
+
 	context = (GnomeMimeContext *) value;
 
 	if (context->mime_type[0] == '#') {
@@ -1502,7 +1326,7 @@ get_key_name (gpointer key, gpointer value, gpointer user_data)
 		return;
 	}
 
-	/* Get name from key and exit if key is NULL or string is empty */	
+	/* Get name from key and exit if key is NULL or string is empty */
 	name = (char *)key;
 	if (name == NULL || strlen (name) == 0) {
 		return;
@@ -1511,7 +1335,7 @@ get_key_name (gpointer key, gpointer value, gpointer user_data)
 	duplicate = NULL;
 	duplicate = g_list_find_custom ((*list), context->mime_type, (GCompareFunc)strcmp);
 	if (duplicate == NULL) {
-		(*list) = g_list_insert_sorted ((*list), g_strdup(context->mime_type), mime_list_sort);		
+		(*list) = g_list_insert_sorted ((*list), g_strdup(context->mime_type), mime_list_sort);
 	}
 
 }
@@ -1521,7 +1345,7 @@ get_key_name (gpointer key, gpointer value, gpointer user_data)
  *
  * resets the user's mime database to the system defaults.
  */
-void 
+void
 gnome_vfs_mime_reset (void)
 {
 	char *filename;
@@ -1529,7 +1353,7 @@ gnome_vfs_mime_reset (void)
 	filename = g_strconcat (user_mime_dir.dirname, "/user.keys", NULL);
 	unlink (filename);
 	g_free (filename);
-       
+
 	filename = g_strconcat (user_mime_dir.dirname, "/user.mime", NULL);
 	unlink (filename);
 	g_free (filename);
@@ -1557,7 +1381,7 @@ gnome_vfs_mime_registered_mime_type_delete (const char *mime_type)
 /*
  * gnome_vfs_get_registered_mime_types
  *
- *  Return the list containing the name of all 
+ *  Return the list containing the name of all
  *  registrered mime types.
  *  This function is costly in terms of speed.
  */
@@ -1565,7 +1389,7 @@ GList *
 gnome_vfs_get_registered_mime_types (void)
 {
 	GList *type_list = NULL;
-	
+
 	if (!gnome_vfs_mime_inited) {
 		gnome_vfs_mime_init ();
 	}
@@ -1602,7 +1426,7 @@ gnome_vfs_mime_registered_mime_type_list_free (GList *list)
  * @mime_type: 	Mime type to set key for
  * @key: 	The key to set
  * @data: 	The data to set for the key
- * 
+ *
  * This function sets the key data for the registered mime
  * type's hash table.
  *
@@ -1627,7 +1451,7 @@ gnome_vfs_mime_set_registered_type_key (const char *mime_type, const char *key,
  * @mime_type: a mime type.
  * @key: A key to lookup for the given mime-type
  *
- * This function retrieves the value associated with @key in 
+ * This function retrieves the value associated with @key in
  * the given GnomeMimeContext.  The string is private, you
  * should not free the result.
  */
@@ -1637,8 +1461,8 @@ gnome_vfs_mime_get_registered_mime_type_key (const char *mime_type, const char *
 	if (!gnome_vfs_mime_inited)
 		gnome_vfs_mime_init ();
 
-	return get_value_real (mime_type, key, 
-			       registered_types_user, 
+	return get_value_real (mime_type, key,
+			       registered_types_user,
 			       registered_types);
 }
 
@@ -1647,7 +1471,7 @@ static DIR *
 ensure_user_directory_exist (void)
 {
 	DIR *dir;
-	
+
 	if (stat (user_mime_dir.dirname, &user_mime_dir.s) != -1)
 		user_mime_dir.valid = TRUE;
 	else
@@ -1655,7 +1479,7 @@ ensure_user_directory_exist (void)
 
 	dir = NULL;
 	dir = opendir (user_mime_dir.dirname);
-	if (dir == NULL){
+	if (dir == NULL) {
 		int result;
 
 		result = mkdir (user_mime_dir.dirname, S_IRWXU );
@@ -1664,90 +1488,62 @@ ensure_user_directory_exist (void)
 			return NULL;
 		}
 		dir = opendir (user_mime_dir.dirname);
-		if (dir == NULL){
+		if (dir == NULL) {
 			user_mime_dir.valid = FALSE;
 		}
-	} 
+	}
 
 	return dir;
 }
 
 
-void  write_back_mime_user_file_context_callback (gpointer key,
-						  gpointer value,
-						  gpointer user_data);
-void  write_back_mime_user_file_context_callback (gpointer key,
-						  gpointer value,
-						  gpointer user_data)
+static void
+write_back_mime_user_file_context_callback (char *key_data,
+					    char *value_data,
+					    FILE *file) 
 {
-	char *key_data, *value_data;
-	FILE *file;
-
-	key_data = (char *) key;
-	value_data = (char *) value;
-	file = (FILE *) user_data;
-
 	fprintf (file, "\t%s: %s\n", key_data, value_data);
 }
-
-void  write_back_mime_user_file_callback (gpointer key,
-					  gpointer value,
-					  gpointer user_data);
-void  write_back_mime_user_file_callback (gpointer key,
-					  gpointer value,
-					  gpointer user_data)
+static void
+write_back_mime_user_file_callback (char		*mime_type,
+				    GnomeMimeContext	*context,
+				    FILE		*file)
 {
-	GnomeMimeContext *context;
-	char *mime_type;
-	FILE *file;
-
-	context = (GnomeMimeContext *) value;
-	mime_type = (char *)key;
-	file = (FILE *) user_data;
-
-	/* print the mime type */
 	fprintf (file, "%s\n", mime_type);
-
-	/* print the mime type keys */
-	g_hash_table_foreach (context->keys, 
-			      write_back_mime_user_file_context_callback,
-			      file);
-	
-	/* add a blank kine between mime types */
+	g_hash_table_foreach (context->keys,
+		(GHFunc) write_back_mime_user_file_context_callback,
+		file);
 	fprintf (file, "\n");
-
 }
 
 static GnomeVFSResult
-write_back_mime_user_file (void) 
+write_back_mime_user_file (void)
 {
 	DIR *dir;
 	FILE *file;
 	char *filename;
-	
+
 	dir = ensure_user_directory_exist ();
 	if (dir == NULL) {
 		return gnome_vfs_result_from_errno ();
 	}
 
-	
-	if (!user_mime_dir.system_dir){
+	if (!user_mime_dir.system_dir) {
 		filename = g_strconcat (user_mime_dir.dirname, "/user.mime", NULL);
 
         	remove (filename);
 		file = fopen (filename, "w");
-		if (file == NULL) {
+		if (file == NULL)
 			return gnome_vfs_result_from_errno ();
-		}
 
 		fprintf (file,
 			 "# This file was autogenerated by gnome-vfs-mime-info.\n"
 			 "# Do not edit by hand.\n");
 
-		g_hash_table_foreach (registered_types_user, 
-				      write_back_mime_user_file_callback,
-				      file);
-		
+		g_hash_table_foreach (registered_types_user,
+			(GHFunc) write_back_mime_user_file_callback,
+			file);
+
 		/* Cleanup file */
 		fclose (file);
 		g_free (filename);
@@ -1756,65 +1552,39 @@ write_back_mime_user_file (void)
 	return GNOME_VFS_OK;
 }
 
-void  write_back_keys_user_file_context_callback (gpointer key,
-						  gpointer value,
-						  gpointer user_data);
-void  write_back_keys_user_file_context_callback (gpointer key,
-						  gpointer value,
-						  gpointer user_data)
+static void
+write_back_keys_user_file_context_callback (char *key_data,
+					    char *value_data,
+					    FILE *file)
 {
-	char *key_data, *value_data;
-	FILE *file;
-
-	key_data = (char *) key;
-	value_data = (char *) value;
-	file = (FILE *) user_data;
-
 	fprintf (file, "\t%s=%s\n", key_data, value_data);
-
 }
+static void
+write_back_keys_user_file_callback (char	     *mime_type,
+				    GnomeMimeContext *context,
+				    FILE	     *file)
 
-void  write_back_keys_user_file_callback (gpointer key,
-					  gpointer value,
-					  gpointer user_data);
-void  write_back_keys_user_file_callback (gpointer key,
-					  gpointer value,
-					  gpointer user_data)
 {
-	GnomeMimeContext *context;
-	char *mime_type;
-	FILE *file;
-
-	context = (GnomeMimeContext *) value;
-	mime_type = (char *)key;
-	file = (FILE *) user_data;
-
-	/* print the mime type */
 	fprintf (file, "%s\n", mime_type);
-
-	/* print the mime type keys */
-	g_hash_table_foreach (context->keys, 
-			      write_back_keys_user_file_context_callback,
-			      file);
-	
-	/* add a blank kine between mime types */
+	g_hash_table_foreach (context->keys,
+		(GHFunc) write_back_keys_user_file_context_callback,
+		file);
 	fprintf (file, "\n");
-
 }
 
 static GnomeVFSResult
-write_back_keys_user_file (void) 
+write_back_keys_user_file (void)
 {
 	DIR *dir;
 	FILE *file;
 	char *filename;
-	
+
 	dir = ensure_user_directory_exist ();
 	if (dir == NULL) {
 		return gnome_vfs_result_from_errno ();
 	}
 
-	if (!user_mime_dir.system_dir){
+	if (!user_mime_dir.system_dir) {
 		filename = g_strconcat (user_mime_dir.dirname, "/user.keys", NULL);
 
         	remove (filename);
@@ -1827,27 +1597,14 @@ write_back_keys_user_file (void)
 		fprintf (file, "# this file was autogenerated by gnome-vfs-mime-info.\n"
 			 "# DO NOT EDIT BY HAND\n");
 
-		g_hash_table_foreach (specific_types_user, 
-				      write_back_keys_user_file_callback,
-				      file);
-				
+		g_hash_table_foreach (specific_types_user,
+			(GHFunc) write_back_keys_user_file_callback,
+			file);
+
 		/* Cleanup file */
 		fclose (file);
 		g_free (filename);
 	}
 
-	return GNOME_VFS_OK;	
+	return GNOME_VFS_OK;
 }
-
-
-
-
-
-
-
-
-
-
-
-
-
